header {
    import antlr.collections.AST;
}

/** 
 * Ambigious expression grammar, with Antlr's AST construction.
 * 
 * exp		-> exp (+ | -) | exp |
 *             exp (* | /) |
 *             NUMBER 
 *
 * Presedence rules:  
 *   + and - has lower precedence than * and /.
 * Associativity rules: 
 *   +, -, * and / are left associative.
 *
 * Problems:
 *   1. Left recursion.
 *   2. Maintain precedence (=> precedence cascade).
 *   3. Maintain associativity 
 *         (=> check that recursion direction matches associativity, if not
 *             ast nodes must be corrected by parameters to the productions). 
 *
 * Solutions:
 *   1. Rewrite the grammar.
 *   2. Rewrite the grammar (precedence cascade).
 *   3. Construct the nodes from left to right in an EBNF rule (with an 
 *      implied left to right iteration). Or, rewrite as BNF with lhs as a 
 *      parameter to the rhs recursion.
 *
 */
class AmbigiousExpASTParser extends Parser;

options {
    buildAST = true;
    importVocab = AmbigiousExp;
}

{ 
    private boolean parseFailure = false;

    /** 
     * Notes whether parsing failures has occured, after Antlr's 
     * default error reporting. 
     * Note that this metod works in conjunction with the 
     * defaultErrorHandler option. 
     */
    public void reportError(RecognitionException ex) {
        super.reportError(ex);
        parseFailure = true;
    }

    public static void main(String args[]) throws Exception {
        final AmbigiousExpASTParser parser = 
            new AmbigiousExpASTParser(new AmbigiousExpLexer(System.in));
        parser.exp();
        if (!parser.parseFailure) {
            final AST t = parser.getAST();
            System.out.println(t.toStringList());
            AmbigiousExpWalker w = new AmbigiousExpWalker();
            System.out.println("=> " + w.exp(t));
        }
    }
}

exp     : term ((PLUS^ | MINUS^) term)*
        EOF!
        ;
term    : factor ((MULT^ | DIV^) factor)*
        ;
factor  : NUMBER
        ;


class AmbigiousExpWalker extends TreeParser;

options {
    importVocab = AmbigiousExp; 

}

exp returns[int r] 
{ 
    int a, b; 
    r = 0; 
}
        : #(PLUS a=exp b=exp)  { r = a + b; }
        | #(MINUS a=exp b=exp) { r = a - b; }
        | #(MULT a=exp b=exp)  { r = a * b; }
        | #(DIV a=exp b=exp)   { r = a / b; }
        | i:NUMBER             { r = Integer.parseInt(i.getText()); }
        ;
